/* filter.c
 * S.Coffin	GV Computing 4/93
 *
 * process I/O stream between user and igs
 */

/*  Copyright (c) 1992,1993,1994,1995 by Stephen Coffin.  All rights reserved.
 *
 *  This program is distributed in the hope that it will be useful.
 *  Use and copying of this software and preparation of derivative works
 *  based upon this software are permitted, so long as the following
 *  conditions are met:
 *       o credit to the authors is acknowledged following current
 *         academic behavior
 *       o no fees or compensation are charged for use, copies, or
 *         access to this software
 *	 o no fees or compensation are charged for access to the
 *	   game server for which this software is a client
 *       o this copyright notice is included intact.
 *  This software is made available AS IS, and no warranty is made about 
 *  the software or its performance. 
 * 
 *  Bug descriptions, use reports, comments or suggestions are welcome.
 *  Send them to    scoffin@netcom.com
 */

/* include this line to enable stone clicking sound instead of beep if your
 * machine supports it   =SC
 */
/* #define SOUND */

#include "filter.h"
#include "shared.h"
#include <signal.h>

#include "goprocs.h"

extern game maingame;
extern game stepgame;
extern unsigned char teachflag;
extern unsigned char sgflocalflag, localflag;
extern char *strtok();
extern int filterstatus;
extern unsigned char savedir[];
char prompt_str[30] = "xigc";

#define MAX_NAME_LEN 20
#define MAX_RANK_LEN 10
struct game_desc {
	int game_no; /* game number */
	char white_name[MAX_NAME_LEN];	/* white player name */
	char white_rank[MAX_RANK_LEN];	/* white rank */
	char black_name[MAX_NAME_LEN];	/* black player */
	char black_rank[MAX_RANK_LEN];	/* black rank */
	int white_num_rank; /* Numeric rank of white */
	int black_num_rank; /* Numeric rank of black */
	int move_num;	/* move number */
	int board_size;	/* board size */
	int handicap;	/* handicap */
	float komi;	/* komi */
	int byo_yomi_min;	/* byo yomi minutes */
	char free;	/* 'F' if it is an unrated game */
	char rules;	/* 'I' if IGS rules, 'G' if Ing rules */
	int ob_cnt;	/* observer count */
	struct game_desc *next_desc; /* To make lists */
	char asci_desc [100]; /* Simplifies printing to just copy it here */
};

static unsigned char kbd_str[LINESIZE];
ObsState observeflag = NOOBSERVE;
static unsigned char pipeflag = FALSE;
static int pipepid = 0;
static unsigned char infileflag = FALSE;
int move = (-1);
int stepmove = (-1);
int sgf_action;
int sgf_lookfor;
int sgf_var_begin[200];
int sgf_var_depth;
extern int currentmove;
extern int tersemode, sortwhomode;
static FILE *pipe_stdin, *pipe_stdout, *pipe_stderr;
extern FILE *sgf_file_fp;
extern char sgflocalfile[];
extern int rdwr_popen();
extern void rdwr_pclose();
extern void SetSize();
static int nlflag = FALSE;
int moves_in_a_row = 0;
static int game_info_flag = (-1);
static void who();
extern void Quit();
extern void make_undo_sensitive();
long ots, nts;
static unsigned char loginflag = TRUE;

extern int gameno, handicap;
extern int boardsize;
extern unsigned char player1[], player2[];
extern char rules, ffre;
unsigned char splayer1[20], splayer2[20];
int smove;
float komi=0.0;
int sgf_version;
unsigned char rank1[10], rank2[10];
unsigned char srank1[10], srank2[10];
extern int p1_sec, p2_sec;
extern int p1_captured, p2_captured;
extern int p1_byoyomi, p2_byoyomi;
int sp1_sec, sp2_sec;
int sp1_captured, sp2_captured;
int sp1_byoyomi, sp2_byoyomi;
float sp1_komi=0.0;
int sp1_hcap;
int sgameno;
static unsigned char exitflag = FALSE;
extern int colorflag;
extern int set_cursor_color();
extern int boardchanged, mouseflag;
extern unsigned char statusflag;
unsigned char my_moveflag;
static FILE *ffp;

extern int clickflag;
extern int dbg;
extern void catcher();
extern void score_it();
extern void setMark();
extern void do_refresh(), RefreshHost();
extern void clearit(), child_dead();
unsigned char *SCstrstr();
extern int Xsetup_p(), Xclose_p();

#ifdef SOUND
unsigned char stone[] = {
    0xae, 0xa9, 0x56, 0x2b, 0xcf, 0xa8, 0xba, 0x30, 0x2a, 0x3a,
    0x38, 0x35, 0x48, 0xbf, 0xa8, 0xab, 0xec, 0x34, 0xcd, 0xaf,
    0xce, 0x32, 0x26, 0x33, 0xc8, 0xaf, 0xbc, 0x3c, 0x34, 0x43,
    0x4b, 0x4e, 0xec, 0xe5, 0x44, 0x4f, 0xbf, 0xbe, 0xcc, 0xbf,
    0xbc, 0x7b, 0x4a, 0xfe, 0xcf, 0xd6, 0x52, 0x34, 0x31, 0x58,
    0xcf, 0xe0, 0x55, 0xc7, 0xbb, 0xce, 0xd0, 0xb8, 0xb7, 0xe9,
    0x40, 0x3f, 0x4f, 0xc8, 0xae, 0xad, 0xc6, 0x71, 0xd8, 0xe4,
    0x4e, 0x4d, 0x4a, 0x34, 0x2d, 0x3d, 0x7b, 0xfe, 0x4e, 0x4d,
    0x66, 0x61, 0x45, 0x45, 0xd5, 0xe0, 0x3d, 0x47, 0xc8, 0xbd,
    0xc6, 0xc8, 0xb8, 0xbb, 0x66, 0x4d, 0xdf, 0xca, 0xd1, 0xcf,
    0xc4, 0xc3, 0x7d, 0x40, 0x4a, 0xe9, 0x61, 0x3c, 0x43, 0xca,
    0xbb, 0xe6, 0x3d, 0x4d, 0xda, 0x49, 0x37, 0x5f, 0xbd, 0xd3,
    0x3f, 0x49, 0xcd, 0xc2, 0xc3, 0xcd, 0xeb, 0x74, 0xec, 0xe9,
    0x7d, 0x79, 0xd8, 0xc5, 0xbe, 0xc4, 0x5c, 0x34, 0x2f, 0x38,
    0x44, 0x3d, 0x42, 0xdc, 0xc8, 0xd0, 0xc9, 0xb5, 0xaf, 0xbd,
    0x5a, 0x40, 0x6a, 0xbb, 0xbd, 0x73, 0x45, 0x4b, 0x5c, 0xff,
    0xc6, 0xbe, 0x66, 0x35, 0x32, 0x36, 0x45, 0xdd, 0xcd, 0xe9,
    0x4d, 0x49, 0x52, 0xe3, 0xc9, 0xc3, 0xc7, 0xca, 0xc7, 0xbf,
    0xbd, 0xce, 0x61, 0xf9, 0xe5, 0x5c, 0x3f, 0x3f, 0x51, 0xe8,
    0xdf, 0x56, 0x48, 0x50, 0x63, 0x56, 0x5a, 0x5f, 0x69, 0xea,
    0xd0, 0xd7, 0xe0, 0xd6, 0xc9, 0xc8, 0xdb, 0xdf, 0xe2, 0xe9,
    0xd9, 0xce, 0x7f, 0x3b, 0x37, 0x4a, 0x53, 0x3c, 0x3d, 0x6a,
    0xc6, 0xb8, 0xb6, 0xc5, 0x5b, 0x41, 0x4a, 0xd0, 0xc4, 0xcf,
    0xd2, 0xdd, 0x65, 0x5c, 0xd6, 0xc7, 0xd2, 0x55, 0x45, 0x4c,
    0x72, 0xed, 0x62, 0x5c, 0xea, 0xd9, 0x63, 0x4e, 0x4f, 0x56,
    0x50, 0x54, 0xe7, 0xd8, 0xf2, 0x00
};

sound(s) char *s;
{
    FILE *fp;

    if (fp = fopen("/dev/audio", "w"))
    {
        fputs(s, fp);
        fclose(fp);
    }
	else
		  clickflag = 0;
}
#endif

/* Following code supports sorting of the output of the games command */
/*
 *               games     and
 *               games -R : sorting according to rank of players (descending)
 *               games -r : sorting according to rank of players (ascending)
 *               games -O : sorting according to num of observers (descending)
 *               games -o : sorting according to num of observers (ascending)
 *               games -M : sorting according to num of moves played (descending)
 *               games -m : sorting according to num of moves played (ascending)
 *               games -G : sorting according to game number (descending)
 *               games -g : sorting according to game number (ascending)
 */

static int (*games_comp)() = NULL; /* Chosen comparison function */
static struct game_desc *game_desc_list = NULL;
static int read_games_header = 0;

static int rank_comp_up(game_descp1, game_descp2)
    struct game_desc *game_descp1, *game_descp2;
{
    if (game_descp1->white_num_rank < game_descp2->white_num_rank)
		  return 1;
	if (game_descp1->white_num_rank > game_descp2->white_num_rank)
		  return -1;
    if (game_descp1->black_num_rank < game_descp2->black_num_rank)
		  return 1;
	if (game_descp1->black_num_rank > game_descp2->black_num_rank)
		  return -1;

    return 0;
}

static int rank_comp_down(game_descp1, game_descp2)
    struct game_desc *game_descp1, *game_descp2;
{
	return -rank_comp_up(game_descp1, game_descp2);
}

static int ob_num_comp_up(game_descp1, game_descp2)
    struct game_desc *game_descp1, *game_descp2;
{
	if (game_descp1->ob_cnt < game_descp2->ob_cnt) return 1;
	if (game_descp1->ob_cnt > game_descp2->ob_cnt) return -1;
	return 0;
}

static int ob_num_comp_down(game_descp1, game_descp2)
    struct game_desc *game_descp1, *game_descp2;
{
	return -ob_num_comp_up(game_descp1, game_descp2);
}

static int moves_comp_down(game_descp1, game_descp2)
    struct game_desc *game_descp1, *game_descp2;
{
	if (game_descp1->move_num < game_descp2->move_num) return -1;
	if (game_descp1->move_num > game_descp2->move_num) return 1;
	return 0;
}

static int moves_comp_up(game_descp1, game_descp2)
    struct game_desc *game_descp1, *game_descp2;
{
	return -moves_comp_down(game_descp1, game_descp2);
}

static int game_num_comp_up(game_descp1, game_descp2)
    struct game_desc *game_descp1, *game_descp2;
{
	if (game_descp1->game_no > game_descp2->game_no) return -1;
	if (game_descp1->game_no < game_descp2->game_no) return 1;
	return 0;
}

static int game_num_comp_down(game_descp1, game_descp2)
    struct game_desc *game_descp1, *game_descp2;
{
	return -game_num_comp_up(game_descp1, game_descp2);
}

static int convert_rank(rank)
	char *rank;
{
	int result;

	result = atoi(rank);

	if (result <= 0) {
		/* Must be special */
		if( strcmp(rank, "NR") == 0 || strchr( rank, '?' ) != NULL )
			return 0; /* not ranked */
		else return 40; /* Some tournament winner; give him 9d */
	}
	while (*rank && '0' <= *rank && '9' >= *rank)
		rank++;

	switch (*rank) {
	case 'k':
	case 'K':
		return 32 - result;
	case 'd':
	case 'D':
		return result + 31;
	case 'p':
	case 'P':
		return result + 40;
	default:
		/* Something weird */
		return 0;
	}
}

static int enter_game_desc(game_descp)
	struct game_desc *game_descp;
{
	struct game_desc *new_entryp = (struct game_desc *)
		  malloc(sizeof(struct game_desc));
	struct game_desc **tmp;

	if (new_entryp == NULL)
		  return -1;

	*new_entryp = *game_descp;

	for (tmp = &game_desc_list; *tmp && (*games_comp)(new_entryp, *tmp) < 0;
		 tmp = &((*tmp)->next_desc))
		  ;
	new_entryp->next_desc = *tmp;
	*tmp = new_entryp;
	return 0;
}

static void print_games_list()
{
	struct game_desc *tmp, *freep;

	for (tmp = game_desc_list; tmp != NULL;) {
		freep = tmp;
		putout(tmp->asci_desc, TRUE, TRUE, 101);
		tmp = tmp->next_desc;
		free((char *)freep);
	}
	game_desc_list = NULL;
	games_comp = NULL;
	read_games_header = 0;
	if( tersemode ) putout("", TRUE, FALSE, 101);
}

int setup_filter(exec_stdin, exec_stdout, exec_stderr )
	FILE **exec_stdin, **exec_stdout, **exec_stderr;
{
	unsigned char delim = ' ';
	int i, tpid;
	extern unsigned char cmdline[];

	/* handle signals carefully */
	for( i = 1; i <= SIGUSR2; ++i ) signal( i, catcher );
	signal( SIGINT, SIG_IGN );
	signal( SIGQUIT, SIG_IGN );
	signal( SIGCHLD, child_dead );

	/* delete this line if your system does not support SIGWINCH =SC  */
	signal( SIGWINCH, SIG_IGN );

	/* and exec, using FILE's for i/o  */
	tpid = rdwr_popen( exec_stdout, exec_stdin, exec_stderr, cmdline,
		NULL, delim );

	if( tpid < 0 ) {
	    fprintf( stderr, "FORK FAILED!\n" );
	    return( (-1) );
	}
	else if( dbg ) fprintf( stderr, "Fork succeeds; pid=%d\n", tpid );

	return( tpid );
}


/* this one parses input of interest from the comm channel, and
 * takes appropriate action to place stones on the board
 */

process_input( str )
	unsigned char *str;
{
	int i, ixx, j, len, ret, tt, t, c2, c4;
	float c3;
	int m, code, yloc;
	unsigned char xloc, color, tflag;
	static unsigned char s[LINESIZE];
	unsigned char sn[LINESIZE], sn2[LINESIZE];
	static int beenhere = FALSE;
	unsigned char *ts, *temp_str, *current_str;
	unsigned char p1[20], p2[20], r1[20];
	int x;
	char fformat;

	if( !beenhere ) {
	    beenhere = TRUE;
	    for( j=0; j<LINESIZE; ++j ) s[j] = '\0';
	}

	if( dbg ) {
	    fprintf( stderr, "In process_input: str=>>%s<<\n", str );
	    fflush( stderr );
	}

	for( temp_str = str ;; ) {

	    /* find newline */
	    for( i=0; temp_str[i] != '\n'
			&& temp_str[i] != '\0'; ++i );

	    /* no newline..... */
	    if( temp_str[i] == '\0' ) {

		if( i > 0 ) {
		    strcat( (char *)s, (char *)temp_str );
		}

		/* get code */
		if( !isdigit( s[0] ) ) ret = 0;
		else ret = sscanf( (char *)s, "%d", &code );

		 /* no code, print it */
		if( ret != 1 ) {
		    putout( s, TRUE, FALSE, 1 );
		    for( j=0; j<LINESIZE; ++j ) s[j] = '\0';
		}

		return( 0 );
	    }

	    if( temp_str[i] == '\n' ) {

		len = strlen( (char *)s );

		strncat( (char *)s, (char *)temp_str, i );
		s[i+len+1] = '\0';

		/* get code */
		if( !isdigit( s[0] ) ) ret = 0;
		else {
		    ret = sscanf( (char *)s, "%d", &code );
		}

		/* no code, print it */
		if( ret != 1 ) {
			putout( s, TRUE, TRUE, 2 );
		}

		/* inappropriate code */
		else if( ret == 1 && infileflag &&
			 !iscodefile( (MessageType)code, s ) ) {
		    putout( s, TRUE, TRUE, 22 );
		}
		else {
		    /* find space */
		    for(j=0; s[j] != ' ' && s[j] != '\n' ; ++j );
		    if( s[j] == ' ' ) current_str = &s[++j];
		    else if( s[j] == '\n' )
			current_str = &s[++j];
		    else current_str = s;

		    switch( code ) {
			case HELP:	/* these have a "File" after code */
			case THIST:
			case PROVERB:
			case SAY:
			case LAST:
			case MESSAGE:
			case STORED:
			case SGFCMD:
			case INFO:	/* STORED is under INFO too */
			    if( observeflag == PLAYING && ffre == 'T' ) teachflag = TRUE;
			    if( observeflag && ffre == 'T' &&
				strncmp( (char *)current_str, "Game is titled", 14 ) == 0 )
					break;
			    if( observeflag == OBSERVING &&
			        strncmp( (char *)current_str, "You have marks at", 17 ) == 0 )
			        break;
			    if( !loginflag ) {
			    	loginflag = TRUE;
			    	if( strncmp( (char *)current_str, "Set singlegame to be True", 25 ) == 0 )
			        	putout( "", TRUE, TRUE, 147 );
				}
			case FIL:
			    /* scoring done under "INFO" msg */
			    if( observeflag == PLAYING ) {
				float xx;
				ret = sscanf( (char *)current_str,
				    "Set the komi to %f", &xx );
				if( ret == 1 ) {
				    komi = xx;
				    update_handicap( handicap, komi );
				}
			    }
			    else if( observeflag == SCORING ) {
				ret = sscanf( (char *)current_str,
					"Removing @ %c%d", &xloc, &yloc );
				if( ret == 2 ) {
		        	    xloc = xloc - 'A' + 1;
		        	    if( xloc > 8 ) --xloc;
				    score_it( xloc, yloc );
				    get_prisoners( &p1_captured, &p2_captured );
				    if( !statusflag )
					update_status( gameno, rules, ffre,
					    player2,
					    player1, p2_captured, p1_captured,
					    p2_sec, p1_sec, p2_byoyomi,
					    p1_byoyomi, rank1, rank2 );
				}
			    }
			    if( strncmp( (char *)current_str, "File", 4 ) == 0 ) {
				putout( &current_str[4], TRUE, TRUE, 3 );
				if( infileflag ) infileflag = FALSE;
				else infileflag = TRUE;
			    }
			    else {
			        if( code == INFO && (filterstatus & 1) ) {
				    if( isfiltered( current_str, 1 ) != (-1) ||
					!SCstrstr( (char *)current_str, " bugging ") ) {
			                putout( current_str, TRUE, TRUE, 4 );
				    }
			        }
			        else putout( current_str, TRUE, TRUE, 4 );
			    }
			    break;
			case UNDO:	/* undo */
			    putout( current_str, TRUE, TRUE, 6 );
			    if( tersemode && observeflag == PLAYING )
				putout( "\n", TRUE, FALSE, 66 );
			    move = undo( move );
			    break;
	    		case MOVE:	/* a move */
			    /* game number? */
			    ret = sscanf( (char *)current_str,
				    "Game %d %c: %s (%d %d %d) vs %s (%d %d %d)",
					&t, &fformat, player1, &p1_captured,
					&p1_sec,
					&p1_byoyomi, player2, &p2_captured,
					&p2_sec, &p2_byoyomi );

			    if( strncmp( (char *)current_str, "Game", 4 )== 0 ) {
				if( (fformat != IRULES && fformat != GRULES 
				     && fformat != PRULES ) ) {
				fprintf( stderr, "\rLOCAL: this version of xigc supports only go games.\n" );
				fprintf( stderr, "Toggle client mode off to play or observe other game types\n" );
				goto done;
			    }

			    else {
				rules = fformat;
			    }
			    }

			    if( ret == 10 ) {
				if( t != gameno ) {
				    gameno = t;
				    fprintf( stderr,
					"\rLOCAL: Changing game number to %d\n",
					gameno );
				}

				if( !statusflag ) {
				    /* add penalty prisoners for GOE rules */
				    if( rules == GRULES ) {
					if( p1_byoyomi > 0 )
					    p2_captured += p1_byoyomi;
					if( p2_byoyomi > 0 )
					    p1_captured += p2_byoyomi;
				    }
				    set_prisoners( p1_captured, p2_captured );
				    update_status( gameno, rules, ffre, player2,
					player1, p2_captured, p1_captured,
					p2_sec, p1_sec, p2_byoyomi,
					p1_byoyomi, rank1, rank2 );
				}
				rules = fformat;
			    }

			    /* handicap? */
			    ret = sscanf( (char *)current_str,
				"%d(%c): Handicap %d", &m, &color, &tt );
			    if( ret == 3 ) {
			        sprintf( (char *)sn2, "Handicap %d\n", tt );
				if( !tersemode ) {
				    putout( sn2, TRUE, TRUE, 34 );
				    ++moves_in_a_row;
				}
				move = (-1);
				refreshboard();
				handicap = tt;
				sethandicap( tt, REALMOVE );
				update_handicap( handicap, komi );
				if( !boardchanged && !mouseflag ) {
				    set_cursor_color( move+1 );
				}
				goto done;
			    }

			    /* its a move? */
	            	    ret = sscanf( (char *)current_str, "%d(%c): %c%d",
				&m, &color, &xloc, &yloc );
		            if( ret == 4 ) {
				if( stepmove >= move || stepmove == (-1) )
						stepmove = m;
				if( !boardchanged && !mouseflag )
					currentmove = m;

				if( m > 0 && move > 0 && m != move+1 ) {
					beep();
					fprintf( stderr,
						"\nERROR: Move %d found; expected move %d!\n", m, move+1 );
					fprintf( stderr,
						"ERROR: Move %d was not received from IGS!\n", move+1 );
					fprintf( stderr,
						"ERROR: Board is incorrect; move %d is missing!\n", move+1 );
					fprintf( stderr,
						"ERROR: Use \"Refresh (Host)\" to reload game!\n\n" );
					host_highlight( TRUE );
					beep();
				}
				move = m;
				ots = time( NULL );
	                        if( dbg ) {
		                    fprintf( stderr,
			                "Move=%d, color=%c, xloc=%c, yloc=%d\n",
			                move, color, xloc, yloc );
		                }
		        	ret = xloc - 'A' + 1;
		        	if( ret > 8 ) --ret;
				if ((!handicap &&
				      (move == maingame.stoneNumber)) ||
				      (handicap &&
				       (move + handicap - 1) == maingame.stoneNumber
					 )) {
		        	    if( color == 'W' ) {
		            	        makemove(ret, yloc, move,
						WHITE, REALMOVE,
						!boardchanged && !mouseflag,
						((rules==IRULES)?RULES:GOE_RULES));
				    }
		        	    else if( color == 'B' ) {
		            	        makemove( ret, yloc, move,
						BLACK, REALMOVE,
						!boardchanged && !mouseflag,
						((rules==IRULES)?RULES:GOE_RULES));
				    }
				}
				if( !boardchanged && !mouseflag ) {
				    set_cursor_color( move );
				}
				update_move( current_str );

				if( !tersemode ) {
				    /* print the incoming move on
				     * the output stream....
				     */
				    if( moves_in_a_row )
				        putout( current_str, TRUE, TRUE, 7 );
				    else {
				      if( observeflag != PLAYING )
					  putout( "\n", TRUE, FALSE, 77 );
				      if( move == 0 )
					putout( current_str, TRUE, TRUE, 777 );
				      else
					putout( current_str, TRUE, FALSE, 777 );
				    }
				}
				++moves_in_a_row;
	            	    }

			    /* its a pass ?? */
			    else if( SCstrstr( current_str, "Pass" ) != NULL ) {
				makemove( (-1), (-1), ++move, EMPTY, REALMOVE,
					!boardchanged && !mouseflag,
					((rules==IRULES)?RULES:GOE_RULES));
				if( stepmove >= move-1 || stepmove == (-1) )
						stepmove = move;
				if( !boardchanged && !mouseflag )
					currentmove = m;
				update_move( current_str );
				set_cursor_color( colorflag+1 );
				putout( current_str, TRUE, TRUE, 8 );
				if( tersemode ) putout( "\n", TRUE, TRUE, 88 );
			    }
				
			  done:
			    if( tersemode ) {
					nlflag = RETURN;
				}
		    	break;
			case PROMPT:	/* prompt */
			    if( !tersemode ) {
			        if( moves_in_a_row > 1 )
					fprintf( stderr, "\n" );
				else if( my_moveflag &&
						observeflag == PLAYING )
					fprintf( stderr, "\n" );
			    }
			    if( tersemode ) {
				if( nlflag != FALSE ) {
				    if(  moves_in_a_row <= 0 ) {
				        fprintf( stderr, "\n" );
				    }
				    else if( my_moveflag &&
					     observeflag == PLAYING ) {
				        fprintf( stderr, "\r" );
				    }
				}
			    	nlflag = RETURN;
			    }

			    my_moveflag = FALSE;
			    if( pipepid ) pipe_cleanup( TRUE );
			    sscanf( (char *)current_str, "%d", &c2 );
			    switch( c2 ) {
				case LOGON:
				    fprintf( stderr, "Login: " );
				    break;
				case PASSWORD:
				    fprintf( stderr, "Password: " );
				    exitflag = FALSE;
				    break;
				case PASSWD_NEW:
				   fprintf( stderr, "New Password: " );
				   break;
				case PASSWD_CONFIRM:
				   fprintf( stderr, "Re-enter Password: " );
				   break;
				case WAITING:
					if (read_games_header) {
						print_games_list();
					}
				    if( sgflocalflag ) fprintf( stderr, "%s [sgf]> ", prompt_str );
				    else fprintf( stderr, "%s> ", prompt_str );
				    if( observeflag != NOOBSERVE ) {
				    	 teachflag = FALSE;
				    	 time_highlight( FALSE );
				    }
				    if( observeflag == PLAYING ) make_undo_sensitive( FALSE );
				    observeflag = NOOBSERVE;

				    rules = IRULES;
				    break;
				case NOWPLAYING:
				    /* refresh board if just start playing */
				    if( !exitflag && observeflag != PLAYING ) {
				        get_game_info( gameno );
				        make_undo_sensitive( TRUE );
						RefreshHost( NULL, NULL, NULL );
						time_highlight( FALSE );
				    }
					if (read_games_header) {
						print_games_list();
					}
					if( teachflag ) 
				    	fprintf( stderr, "%s [t%d]> ", prompt_str, gameno );
				    else fprintf( stderr, "%s [p%d]> ", prompt_str, gameno );

				    observeflag = PLAYING;
				    break;
				case NOWSCORING:
					if (read_games_header) {
						print_games_list();
					}
				    fprintf( stderr, "Remove dead group> " );
				    observeflag = SCORING;
				    break;
				case NOWOBSERVING:
					if( observeflag != OBSERVING ) time_highlight( FALSE );
					if (read_games_header) {
						print_games_list();
					}
				    if( tersemode && observeflag == NOOBSERVE )
						fprintf( stderr, "\n" );
				    fprintf( stderr, "%s [o%d]> ", prompt_str, gameno );
				    observeflag = OBSERVING;
				    break;
				default:
					if (read_games_header) {
						print_games_list();
					}
				    putout( current_str, TRUE, TRUE, 9 );
				    observeflag = NOOBSERVE;
				    rules = IRULES;
				    teachflag = FALSE;
				    break;
			    }
			    if( exitflag ) {
				observeflag = NOOBSERVE;
				rules = IRULES;
				if( tersemode ) {
					nlflag = TRUE;
				}
			    }
			    fflush( stderr );
			    moves_in_a_row = 0;
			    break;
			case SHOUT:
			    /* update game number if needed */
			    if( (filterstatus & 1) ) {
				if( isfiltered( current_str, 1 ) != (-1) ) {
			            putout( current_str, TRUE, TRUE, 10 );
				}
			    }
			    else putout( current_str, TRUE, TRUE, 10 );
			    break;
			case LOOK_M:
			case STATUS:	/* screen refresh of current display */
			    /* player (w in first line), rank, captured,
			     * time (sec), byoyomi stones, flag (T if byo),
			     * komi, hcap
			     */
			    ret = sscanf( (char *)current_str,
					"%s %s %d %d %d %c %f %d",
					p1, r1, &t, &tt, &c2, &tflag, &c3, &c4 );
			    if( ret == 8 ) {
				if( tflag == 'F' ) c2 = (-1);
				if( !statusflag ) {
				    clearit();
				    boardchanged = TRUE;
				    local_highlight( TRUE );
				    statusflag = TRUE;
				    strcpy( (char *)splayer1, (char *)p1 );
				    strcpy( (char *)srank1, (char *)r1 );
				    sp1_captured = t;
				    sp1_sec = tt;
				    sp1_byoyomi = c2;
				    sp1_komi = c3;
				    sp1_hcap = c4;
				}
				else {
				    strcpy( (char *)splayer2, (char *)p1 );
				    strcpy( (char *)srank2, (char *)r1 );
				    sp2_captured = t;
				    sp2_sec = tt;
				    sp2_byoyomi = c2;
	    			    update_status( sgameno, ' ', ' ', splayer2,
					splayer1,
					sp2_captured, sp1_captured, sp2_sec,
					sp1_sec, sp2_byoyomi, sp1_byoyomi,
					srank1, srank2 );
				    update_move( NULL );
				}
				break;
			    }
			    else {
			        ret = sscanf( (char *)current_str,
					"%d: %s\n", &x, p1 );
			        if( ret == 2 ) parse_line( x, p1 );
			    }
			    break;
			case YELL:	/* reformat the yell output */
			    if( dbg ) fprintf( stderr, "\n\nYELL >>\n%s\n<<\n",
						current_str );
			    if( strncmp( (char *)current_str, "Changing ", 9 ) == 0 ||
				strncmp( (char *)current_str, "Welcome", 7 ) == 0 ) {
				putout( current_str, TRUE, TRUE, 111 );
			    }
			    else {
			        strcpy( (char *)sn2, (char *)current_str );
			        sscanf( (char *)sn2, "%d:%s:", &c2, splayer1 );
			        for( ixx = strlen((char *)splayer1) - 1;
				     ixx > 0 && splayer1[ixx] == ':'; --ixx ) {
				    splayer1[ixx] = '\0';
			        }
			        for( ; sn2[ixx] != ' '; ++ixx );
			        for( ; sn2[ixx] == ' '; ++ixx );
			        ts = &sn2[ixx];
			        for( ixx = strlen( (char *)ts ) -1;
				    ts[ixx] == '\n'; --ixx )
						    ts[ixx] = '\0';
			        sprintf( (char *)sn, "Yell [%d] from %s: %s",
					c2, splayer1, ts );
			        putout( sn, TRUE, TRUE, 11 );
			    }
			    break;
			case GAMES:	/* this is used for getting the player
					 * ranks, komi, etc
					 */
			    {
				struct game_desc game_desc;
		
			    ret = parse_games( current_str, &game_desc);

			    if( ret >= 10 && game_desc.game_no == gameno ) {
				strcpy( (char *)splayer1, game_desc.white_name );
				strcpy( (char *)splayer2, game_desc.black_name );
				strcpy( (char *)rank1, game_desc.white_rank );
				strcpy( (char *)rank2, game_desc.black_rank );
				komi = game_desc.komi;
				ffre = game_desc.free;
				if( game_desc.board_size != boardsize ) {
				    boardsize = game_desc.board_size;
				    if(boardsize  == 9 || boardsize == 13 || boardsize == 19 )
					SetSize( NULL, NULL, boardsize );
				    initboard( boardsize );
				}
				update_handicap( game_desc.handicap, komi );
				rules = game_desc.rules;
			    }
			    if( game_info_flag == (-1) ) {
					if ( games_comp != NULL && ret > 0) {
						strcpy(game_desc.asci_desc, (char *)current_str);
						if (enter_game_desc(&game_desc) < 0) {
							print_games_list();
							putout( current_str, TRUE, TRUE, 131 );
						}
					} else {
						putout( current_str, TRUE, TRUE, 133 );
						if( games_comp != NULL && read_games_header == 0)
							read_games_header = 1;
					}
			    } else if( ret >= 10 && game_desc.game_no == game_info_flag )
				game_info_flag = (-1);
			    }
			    break;
			case ERROR:
#ifdef SC
			    if( strncmp( (char *)current_str,
					 "You are not consider", 20 ) == 0 )
				break;
			    else
#endif
			    if( strncmp( (char *)current_str,
					 "Game not found", 14 ) == 0 )
				game_info_flag = (-1);
			    putout( current_str, TRUE, TRUE, 132 );
			    break;
            case VERSION:
                putout( current_str, TRUE, TRUE, 145 );
                putout( "\n", TRUE, TRUE, 146 );
				sscanf( (char *)current_str, "%s", prompt_str );
				for( ixx = 0; ixx < (int)strlen((char *)prompt_str); ++ixx )
						prompt_str[ixx] = tolower( prompt_str[ixx] );
			    /*XXX toggle singlegame on */
	    		if( strncmp( prompt_str, "igs", 3 ) == 0 ) {
	    			fprintf( ffp, "toggle singlegame on\n" );
	    			fflush( ffp );
	    			loginflag = FALSE;
			    }
			    break;
			case UNKNOWN:	/* strip the code from these, but */
			case BOARD:	/* otherwise pass them untouched */
			case DOWN:
			case TRANS:
			case LOAD:
			case OBSERVE:
			case REFRESH:
			case SAVED:
			case SCORE:
			case SHOW:	/* what is this? */
			case TIM:
			case AUTOMAT:
			case AUTOASK:
			case TTT_BOARD:
			case CHOICES:
			case TELLRET:
			case USER:
			case EMOTE:
			case EMOTETO:
			case PONG:

			    if( strncmp( (char *)current_str, "File", 4 ) == 0 ) {
				putout( &current_str[4], TRUE, TRUE, 12 );
				if( infileflag ) infileflag = FALSE;
				else infileflag = TRUE;
			    }
			    else {
				putout( current_str, TRUE, TRUE, 134 );
			    }
		            break;
			case TELL:
			    if( (filterstatus & 2) ) {
				if( isfiltered( current_str, 2 ) == (-1) ) {
			            putout( current_str, TRUE, TRUE, 134 );
				}
			    }
			    else putout( current_str, TRUE, TRUE, 134 );
		            break;
			case WHO:
			    if( sortwhomode ) who( current_str );
			    else putout( current_str, TRUE, TRUE, 140 );
			    break;

			case KIBITZ:
			    /* We massage the input in order to merge:
			     *
			     * Kibitz jansteen[3d] : Game kato - cho [3]
			     *  wow!
			     *
			     * to: "jansteen[3d]: wow!\n"
			     *
			     * thanks to jansteen@cwi.nl for providing this
			     * code fragment
			     */

			    /*
			     * Remove some header and the trailer stuff
			     */
			    {
			    unsigned char *s = current_str;
			    unsigned char *t;

			    if( !strncmp((char *)s, "Kibitz ", strlen("Kibitz "))) {
				    s += strlen( "Kibitz " );
				    t = (unsigned char *)strchr( (char *)s, ':' );
				    if( t != (unsigned char *)0 ) *(t+1) = 0;
				    putout( "Kibitz from ", FALSE, FALSE, 24 );
				    putout( s, TRUE, FALSE, 14 );
			    }
			    else if( !strncmp((char *)s, "Chatter ", strlen("Chatter "))) {
				    s += strlen( "Chatter " );
				    t = (unsigned char *)strchr( (char *)s, ':' );
				    if( t != (unsigned char *)0 ) *(t+1) = 0;
				    putout( "Chatter from ", FALSE, FALSE, 24 );
				    putout( s, TRUE, FALSE, 14 );
			    }
			    else {
				/*
				 * Just skip one space
				 */
				putout( s+1, TRUE, TRUE, 15 );
			    }
			    }
		            break;
			case BEEP:
			    if( dbg ){
				fprintf( stderr, "BEEP\n" );
				fflush( stderr );
			    }
#ifdef SOUND
				if (clickflag) {
					sound(stone);
					if (!tersemode)
						  ++moves_in_a_row;
			    } else
#endif
				if( tersemode )
				beep();
			    else {
				beep();
				++moves_in_a_row;
			    }
			    break;
			case REVIEW:	/* reviewing a SGF position */
			    break;
			case CLIVRFY:	/* special command echo-- ignore it! */
			    break;
			default:	/* print code if it is unexpected */
		    	    sprintf( (char *)sn, "%d %s", code, current_str );
			    putout( sn, TRUE, TRUE, 18 );
		            break;
		    }
		}
	        
		/* reset active string */
			for( j=0; j<LINESIZE; ++j ) s[j] = '\0';
			temp_str = &temp_str[i+1];
			continue;
	    }
	}
}

int putout( xtr, extraflag, newflag, cc )
	unsigned char *xtr;
	int extraflag, newflag;
	int cc;
{
	int ret;

	if( dbg ) fprintf( stderr, "PUTOUT %d\n", cc );

	/* send to pipe? */
	if( pipeflag ) {
	    if( dbg ) fprintf( stderr, "PIPING >>%s<<=\n", xtr );
	    if( !tersemode || (nlflag && extraflag) )
			fprintf( pipe_stdin, "\n" );
	    ret = strlen( (char *)xtr ) - 1;
	    if( xtr[ret] == '\r' ) xtr[ret] = '\0';
	    fprintf( pipe_stdin, "%s", xtr );
	    if( !tersemode ) fprintf( pipe_stdin, "\n" );
	    fflush( pipe_stdin );
	    nlflag = newflag;
	}

	/* send to screen */
	else if( tersemode ) {
	    if( nlflag == RETURN || !extraflag ) {
	        if( cc != 1 ) {
		    	fprintf( stderr, "\r" );
		    	nlflag = FALSE;
	        }
	    }
	    else {
			if( nlflag || cc == 101 ) fprintf( stderr, "\n" );
	        nlflag = newflag;
	    }
	    fprintf( stderr, "%s", xtr );
	    fflush( stderr );
	}
	else {
		fprintf( stderr, "%s", xtr );
	        if( newflag ) fprintf( stderr, "\n" );
	    fflush( stderr );
	}

	return( 0 );
}


int oboff( flag )
	int flag;
{
	if( observeflag == NOOBSERVE ) {
	    if( flag ) {
		fprintf( stderr, "\rLOCAL: Not currenty observing\n" );
		fflush( stderr );
	    }
	    fprintf( ffp, "\n" );
	}
	else {
	    fprintf( ffp, "unobserve\n" );
	    gameno = (-1);
	}
	fflush( ffp );
	sleep( 2 );
	return( 0 );
}

/* examine user input from keyboard for commands of local interest
 * then send it to host
 */
process_kbd( str, fp )
	unsigned char *str;
	FILE *fp;
{
	int i, k, t, len, ret;
	unsigned char s[40];
	char *ss;
	unsigned char pipecmd[LINESIZE];
	int match_len;
	extern unsigned char keyecho;

	ffp = fp;

	if( dbg ) fprintf( stderr, "In process_kbd:  >>%s<<", str );
	if( tersemode ) {
		nlflag = FALSE;
	}

	for( i=0; str[i] != '\0' && str[i] != '\n'; ++i )
		kbd_str[i] = tolower( str[i] );
	kbd_str[i] = '\0';

	/* find a pipe symbol... */
/*XXX  This does not work....  don't even think about it  =SC
	for( i = 0; kbd_str[i] != '\0' && kbd_str[i] != '|'; ++i );
	len = strlen( (char *)kbd_str );
	if( i < len ) {
	    for( k = i-1; k > 0 && isspace(kbd_str[k]); --k );
	    kbd_str[k+1] = '\0';
	    for( t=i+1; isspace( str[t] ); ++t );
	    strcpy( (char *)pipecmd, (char *)&str[t] );
	    str[i] = '\n';
	    str[i+1] = '\0';
	    i = strlen( (char *)pipecmd );
	    if( pipecmd[i-1] == '\n' ) pipecmd[i-1] = '\0';
	    setup_pipe( pipecmd );
	}
XXX*/

	if( dbg ) fprintf( stderr, "kbd_str=>>%s<<\n", kbd_str );

	if( strncmp( (char *)kbd_str, "ton", 3 ) == 0 ) {
	    tersemode = TRUE;
	    fprintf( stderr, "\rLOCAL:  terse mode ON\n" );
	    fflush( stderr );
	    return( 0 );
	}
	else if( strncmp( (char *)kbd_str, "toff", 4 ) == 0 ) {
	    tersemode = FALSE;
	    fprintf( stderr, "\rLOCAL:  terse mode OFF\n" );
	    fflush( stderr );
	    return( 0 );
	}

#ifdef SC
	if( strncmp( (char *)kbd_str, "dbg", 3 ) == 0 ) {
	    if( dbg ) {
		fprintf( stderr, "\rLOCAL:  debug mode OFF\n" );
		dbg = FALSE;
	    }
	    else {
		fprintf( stderr, "\rLOCAL:  debug mode ON\n" );
		dbg = TRUE;
	    }
	    fprintf( fp, "\n" );
	    fflush( fp );
	    fflush( stderr );
	    return( 0 );
	}
#endif

	/* add a name to the friends list */
	if( match( (char *)kbd_str, "addfriend" ) >= 4 ) {
		strtok( (char *)str, " 	" );
		ss = strtok( NULL, " 	" );
		dofriend( ss );
	        fprintf( fp, "\n" );
	        fflush( fp );
	        fflush( stderr );
	        return( 0 );
	}
	/* remove name from the friends list */
	if( match( (char *)kbd_str, "removename" ) >= 7 ) {
		strtok( (char *)str, " 	" );
		ss = strtok( NULL, " 	" );
		doremovelist( ss );
	        fprintf( fp, "\n" );
	        fflush( fp );
	        fflush( stderr );
	        return( 0 );
	}
	/* add a name to the kill list */
	if( match( (char *)kbd_str, "addkill" ) >= 4 ) {
		strtok( (char *)str, " 	" );
		ss = strtok( NULL, " 	" );
		dokill( ss );
	        fprintf( fp, "\n" );
	        fflush( fp );
	        fflush( stderr );
	        return( 0 );
	}

	/* toggle kill list */
	if( match( (char *)kbd_str, "togk" ) >= 4 ) {
		/* turn off */
		if( filterstatus & 2 ) {
			filterstatus &= ~2;
			fprintf( stderr, "LOCAL: Disable kill list\n" );
		}
		/* turn on */
		else {
			filterstatus |= 2;
			fprintf( stderr, "LOCAL: Enable kill list\n" );
		}
	        fprintf( fp, "\n" );
	        fflush( fp );
	        fflush( stderr );
	        return( 0 );
	}
	/* toggle friends list */
	if( match( (char *)kbd_str, "togf" ) >= 4 ) {
		/* turn off */
		if( filterstatus & 1 ) {
			filterstatus &= ~1;
			fprintf( stderr, "LOCAL: Disable friends list\n" );
		}
		/* turn on */
		else {
			filterstatus |= 1;
			fprintf( stderr, "LOCAL: Enable friends list\n" );
		}
	        fprintf( fp, "\n" );
	        fflush( fp );
	        fflush( stderr );
	        return( 0 );
	}

			
	if( match( (char *)kbd_str, "gameno" ) >= 5 ) {
		sscanf( (char *)kbd_str, "gameno %d", &gameno );
		fprintf( stderr, "\rLOCAL:  Set game number to %d\n", gameno );
		fflush( stderr );
	        fprintf( fp, "\n" );
	        fflush( fp );
		return( 0 );
	}
	if ((match_len = match((char *)kbd_str, "games")) >= 2) {
		char *cp = (char *)&kbd_str[match_len];

		if( games_comp != NULL ) return 0;
		read_games_header = 0;
		while (*cp && *cp == ' ')
			  cp++;
		if (*cp == '\0') {
			/* Sort according to rank */
			games_comp = rank_comp_up;
		} else if (*cp == '-') {
			cp = strchr((char *)str, '-'); /* Need the original char */
			switch (*(cp+1)) {
			case 'R':
				games_comp = rank_comp_up;
				break;
			case 'r':
				games_comp = rank_comp_down;
				break;
			case 'O':
				games_comp = ob_num_comp_down;
				break;
			case 'o':
				/* Sort according to number of observers */
				games_comp = ob_num_comp_up;
				break;
			case 'M':
				/* Sort according to number of moves */
				games_comp = moves_comp_up;
				break;
			case 'm':
				/* Sort according to number of moves */
				games_comp = moves_comp_down;
				break;
			case 'G':
				/* Sort according to game no */
				games_comp = game_num_comp_down;
				break;
			case 'g':
				/* Sort according to game no */
				games_comp = game_num_comp_up;
				break;
			default:
				/* No sorting */
				fprintf(fp, "%s\n", (char *)kbd_str );
				fflush(fp);
				return 0;
			}
		} else {
			/* No sorting */
			fprintf(fp, "%s\n", (char *)kbd_str );
			fflush(fp);
			return 0;
		}
		fprintf(fp , "games\n");
		fflush(fp);
		return 0;
	}

	if( match( (char *)kbd_str, "load" ) >= 3 ) {
	    if( observeflag == SCORING || observeflag == PLAYING ) {
		fprintf( stderr, "\rLOCAL: cannot load while playing\n" );
	        fprintf( fp, "\n" );
	        fflush( fp );
		return( 0 );
	    }
	    if( observeflag == OBSERVING ) {
		oboff( FALSE );
	    }

	    fprintf( fp, (char *)str );
	    fflush( fp );
	    return( 0 );
	}

#ifndef SC
	if( match( (char *)kbd_str, "alias" ) >= 3 ) {
		fprintf( stderr, "\rLOCAL: this version of xigc does not support aliases\n" );
	        fprintf( fp, "\n" );
	        fflush( fp );
		return( 0 );
	}
#endif

	if( match( (char *)kbd_str, "look" ) >= 3 ) {
	    if( observeflag == SCORING ) {
		fprintf( stderr, "\rLOCAL: cannot look while scoring\n" );
	        fprintf( fp, "\n" );
	        fflush( fp );
		return( 0 );
	    }
	    else {
	        statusflag = FALSE;
	        fprintf( fp, (char *)str );
	        fflush( fp );
	        return( 0 );
	    }
	}

	if( match( (char *)kbd_str, "status" ) >= 5 ) {
	    if( observeflag == SCORING ) {
		fprintf( stderr, "\rLOCAL: cannot get status while scoring\n" );
	        fprintf( fp, "\n" );
	        fflush( fp );
		return( 0 );
	    }
	    else {
	        sscanf( (char *)kbd_str, "%*s %d", &sgameno );
	        statusflag = FALSE;
	        fprintf( fp, (char *)str );
	        fflush( fp );
	        return( 0 );
	    }
	}
	else if( ((i=match( (char *)kbd_str, "moves" )) >= 2 &&
		 kbd_str[i] != ' ' ) ||
		 ((i=match( (char *)kbd_str, "refresh" )) >= 3 &&
		 kbd_str[i] != ' ') ) {
	    if( gameno < 0 ) {
		fprintf( stderr, "\rLOCAL: no current game to refresh\n" );
		fflush( stderr );
	        fprintf( fp, "\n" );
	        fflush( fp );
	    }
	    else {
	        move = stepmove = (-1);
		currentmove = (-1);
	        refreshboard();
	        fprintf( fp, "moves %d\n", gameno );
	        fflush( fp );
		host_highlight( FALSE );
	    }
	    return( 0 );
	}
	else if( ((i=match( (char *)kbd_str, "moves" )) >= 2 &&
		 kbd_str[i] == ' ' ) ||
		 match( (char *)kbd_str, "refresh" ) >= 3 ) {

		i = sscanf( (char *)kbd_str, "%*s %d", &ret );
		if( i < 1 ) {
	            fprintf( fp, (char *)str );
	            fflush( fp );
		    return( 0 );
		}
		if( observeflag != NOOBSERVE && ret != gameno ) {
		    fprintf( stderr,
			"\rLOCAL: Can't get moves of another game while observing or playing;\n" );
		    fprintf( stderr, "Use \"status\" instead\n" );
		    fflush( stderr );
	            fprintf( fp, "\n" );
	            fflush( fp );
		    return( 0 );
		}
		else {
	            move = stepmove = (-1);
		    currentmove = (-1);
		    gameno = ret;
		    host_highlight( FALSE );
	            refreshboard();
	            fprintf( fp, (char *)str );
	            fflush( fp );
	            return( 0 );
		}
	}
	else if( match( (char *)kbd_str, "refresh" ) >= 3 ) {
	        move = stepmove = (-1);
		currentmove = (-1);
		i = sscanf( (char *)kbd_str, "%*s %d", &t );
		if( i < 1 ) {
	            fprintf( fp, (char *)str );
	            fflush( fp );
		    return( 0 );
		}
	        refreshboard();
		host_highlight( FALSE );
	        fprintf( fp, "moves %d\n", t );
	        fflush( fp );
		return( 0 );
	}
	else if( match( (char *)kbd_str, "unobserve" ) >= 3 ) {
	    gameno = -1;
	    host_highlight( FALSE );
	    return( oboff( TRUE ) );
	}
	else if( strncmp( (char *)kbd_str, "observe ", 8 ) == 0 ||
		 strncmp( (char *)kbd_str, "ob ", 3 ) == 0 ) {
	    if( sgflocalflag ) {
		fprintf( stderr, "LOCAL: Cannot observe while local SGF file is open\n" );
	        fprintf( fp, "\n" );
	        fflush( fp );
		return( 0 );
	    }
	    if( sscanf( (char *)kbd_str, "observe %d", &t ) != 1 )
		sscanf( (char *)kbd_str, "ob %d", &t );
	    if( observeflag != NOOBSERVE ) {
		oboff( FALSE );
	        fflush( stderr );
	        fprintf( fp, "\n" );
	        fflush( fp );
		if( observeflag == PLAYING || observeflag == SCORING )
				return( 0 );
	    }

	    gameno = t;
	    move = (-1);
	    host_highlight( FALSE );
	    refreshboard();
	    get_game_info( gameno );
	    fprintf( fp, "observe %d\n", gameno );
	    fflush( fp );
	    sleep( 2 );
	    fprintf( fp, "moves %d\n", gameno );
	    fflush( fp );
	    return( 0 );
	}
	else if( strncmp( (char *)kbd_str, "observeoff", 10 ) == 0 ||
		 strncmp( (char *)kbd_str, "obo", 3 ) == 0 ) {
	    host_highlight( FALSE );
	    return( oboff( TRUE ) );
	}
	else if( (i=match( (char *)kbd_str, "all" )) >= 2 &&
		 (int)strlen( (char *)kbd_str ) <= i ) {
	    if( observeflag != NOOBSERVE ) {
		fprintf( fp, "all %d\n", gameno );
		fflush( fp );
		return( 0 );
	    }
	}
	else if( match( (char *)kbd_str, "exit" ) >= 2 ||
		 match( (char *)kbd_str, "quit" ) >= 2 ) {
		observeflag = NOOBSERVE;
		rules = IRULES;
		exitflag = TRUE;
	}
	else if( match( (char *)kbd_str, "match" ) >= 3 ||
		 match( (char *)kbd_str, "pmatch" ) >= 3 ||
		 match( (char *)kbd_str, "tmatch" ) >= 3 ||
		 match( (char *)kbd_str, "tgmatch" ) >= 3 ||
		 match( (char *)kbd_str, "goematch" ) >= 3 ||
		 match( (char *)kbd_str, "automatch" ) >= 3 ||
		 match( (char *)kbd_str, "team" ) >= 3 || 
		 match( (char *)kbd_str, "teach" ) >= 3 ||
		 match( (char *)kbd_str, "gteach" ) >= 3 ) {

	    if( sgflocalflag ) {
		sgflocalflag = FALSE;
		fclose( sgf_file_fp );
		fprintf( stderr, "LOCAL: Ending sgf file %s\n", sgflocalfile );
	    }

	    if( observeflag == OBSERVING ) {
	        fprintf( fp, "observe %d\n", gameno );
	        fflush( fp );
	    }
	    else if( observeflag != NOOBSERVE ) {
		fprintf( stderr, "\rLOCAL: Already playing!\n" );
		fflush( stderr );
	        fprintf( fp, "\n" );
	        fflush( fp );
		return( 0 );
	    }
		if( match( (char *)kbd_str, "teach" ) >= 3 ||
			match( (char *)kbd_str, "gteach" ) >= 3 ) {
				teachflag = TRUE;
		}
	}
	else if( match( (char *)kbd_str, "forward" ) >= 3 ) {
	    fprintf( fp, "forward\n" );
	    fflush( fp );
	    putout( "", TRUE, FALSE, 123 );
	    return( 0 );
	}
	else if( match( (char *)kbd_str, "sgfload" ) >= 4 ) {
	    if( sgflocalflag ) {
		fprintf( stderr, "LOCAL: Cannot load multiple sgf files!\n" );
		return( 0 );
	    }
	    if( observeflag == SCORING || observeflag == PLAYING ) {
		fprintf( stderr, "\rLOCAL: cannot load sgf while playing\n" );
	        fprintf( fp, "\n" );
	        fflush( fp );
		return( 0 );
	    }
	    if( observeflag == OBSERVING ) {
		oboff( FALSE );
	    }

	    ret = sscanf( (char *)str, "%*s %s", sgflocalfile );
	    if( ret != 1 ) {
		fprintf( stderr, "LOCAL: must give filename arg to sgfload command\n" );
		return( 0 );
	    }

	    sgf_file_fp = fopen( (char *)sgflocalfile, "r" );
	    if( sgf_file_fp == NULL ) {
		sprintf( (char *)str, "%s/%s", savedir, sgflocalfile );
	        sgf_file_fp = fopen( (char *)str, "r" );
		if( sgf_file_fp == NULL ) {
		    fprintf( stderr, "LOCAL: cannot open file %s\n", sgflocalfile );
		    return( 0 );
		}
	    }

	    sgf_var_depth = 0;
	    sgf_clear_info();
	    clearit();
	    if( boardsize != 19 ) {
		  boardsize = 19;
		  SetSize( NULL, NULL, boardsize );
	    }
	    initboard( boardsize );
	    local_highlight( FALSE );
	    boardchanged = FALSE;
	    statusflag = FALSE;
	    process_local_sgf_line( 0 );
			
	    fprintf( stderr, "LOCAL: Loading sgf file %s\n", sgflocalfile );
	    fprintf( stderr, "LOCAL: Use \"Step Forward\" to review game\n" );
	    sgflocalflag = TRUE;
	    return( 0 );
	}
	else if( match( (char *)kbd_str, "sgfdone" ) >= 4 ) {
	    if( !sgflocalflag ) {
	        fprintf( stderr, "LOCAL: No sgf file open!\n" );
	    }
	    else {
	        sgflocalflag = FALSE;
	        fclose( sgf_file_fp );
	        fprintf( stderr, "LOCAL: Ending sgf file %s\n", sgflocalfile );
	    }
	    if( !localflag ) {
		fprintf( fp, "\n" );
		fflush( fp );
	    }
	    return( 0 );
	}

	fprintf( fp, "%s", (char *)str );
	fflush( fp );
	return( 0 );
}

/* signal handler to clean up on disasters */
void catcher( sig, code )
	int sig, code;
{
	if( dbg || sig != SIGCHLD )
		fprintf( stderr, "\nSignal catcher [sig=%d code=%d]\n",
			sig, code );

	rdwr_pclose( NULL, NULL, NULL );
	exit( code );
}

int undo( m )
	int m;
{
	if (stepgame.stoneNumber >= Maxstones) {
	  backup(&maingame,maingame.stoneNumber,DRAWIT,((rules==IRULES)?RULES:GOE_RULES));
	  backupend(&maingame,maingame.stoneNumber - 1,DRAWIT,0);
	} else {
	  backup(&maingame,maingame.stoneNumber,NODRAWIT,((rules==IRULES)?RULES:GOE_RULES));
	  backupend(&maingame,maingame.stoneNumber - 1,NODRAWIT,0);
	}
	return( m-2 );
}

/* Find the first occurrence of s2 in s1  */
unsigned char *SCstrstr( s1, s2 )
	unsigned char *s1, *s2;
{
	unsigned char c, sc;
	int len;

	if( (c = *s2++) != 0 ) {
	    len = strlen((char *)s2);
	    do {
		do {
		    if ((sc = *s1++) == 0) return (NULL);
		} while (sc != c);
	    } while (strncmp((char *)s1, (char *)s2, len) != 0);
	    s1--;
	}
	return( s1 );
}

parse_line( line, str )
	int line;
	unsigned char *str;
{
	int xloc, yloc;
	int len;
	int i;
	unsigned char c;

	len = strlen( (char *)str );
	while( str[len-1] == '\n' || str[len-1] == '\r' ) --len;
	str[len] = '\0';

	if( dbg )
	    fprintf( stderr, "In parse_line:  line=%d, str=>>%s<<\n",
					line, str );

	xloc = line + 1;

	for( i = 0; i < len; ++i ) {
	    yloc = len - i;
	    switch( str[i] ) {
		case '0':		/* black stone */
		    drawStone( xloc, yloc, BLACK, FALSE );
		    break;
		case '1':		/* white stone */
		    drawStone( xloc, yloc, WHITE, FALSE );
		    break;
		case '4':		/* White territory */
		    setMark( xloc, yloc, 'w' );
		    update_prisoners( 0, 1 );
		    break;
		case '5':		/* black territory */
		    setMark( xloc, yloc, 'b' );
		    update_prisoners( 1, 0 );
		    break;
		case '2':		/* empty point */
		case '3':		/* dame */
		case '6':		/* starpoint */
		case '9':
		    break;
		case '7':		/* counted */
		    setMark( xloc, yloc, 'c' );
		    break;
		case '8':		/* mark */
		    setMark( xloc, yloc, 'x' );
		    break;
		default:
		    fprintf( stderr,
			"ERROR IN PARSE_LINE: bad code = 0x%x\n", str[i] );
		    break;
	    }
	}
	if( xloc >= len ) {
	    if( !statusflag ) get_prisoners( &sp2_captured, &sp1_captured );
	    /*XXX  include komi in final point count at game end  ?? */
	    update_status( sgameno, ' ', ' ', splayer2, splayer1,
			sp2_captured,
			sp1_captured, sp2_sec, sp1_sec, sp2_byoyomi,
			sp1_byoyomi, srank1, srank2 );
	    update_handicap( sp1_hcap, sp1_komi );
	}
}

int setup_pipe( cmd )
	unsigned char *cmd;
{
	unsigned char delim = ' ';

	if( dbg ) fprintf( stderr, "in setup_pipe  cmd=>>%s<<\n", cmd );
	if( pipeflag ) {
	    fprintf( stderr, "pipe already active; cannot pipe twice\n" );
	    return( (-1) );
	}
	
	pipepid = rdwr_popen( &pipe_stdout, &pipe_stdin, &pipe_stderr,
			cmd, NULL, delim );
	Xsetup_p( pipe_stdin, pipe_stdout, pipe_stderr );
	if( dbg ) fprintf( stderr, "in setup_pipe[pid=%d]  cmd=>>%s<<\n",
			pipepid, cmd );

	if( pipepid < 0 ) {
	    fprintf( stderr, "Pipe open failed  [ret=%d]\n", pipepid );
	    pipepid = 0;
	    pipeflag = FALSE;
	    return( (-1) );
	}
	else pipeflag = TRUE;

	return( pipepid );
}

pipe_cleanup( flag )
	int flag;	/* true if wait, false if already waited */
{
	int ret;

	pipeflag = FALSE;
	if( TRUE ) {
	    fprintf( stderr, "in pipe_cleanup[pid=%d, flag=%d]\n",
			pipepid, flag );
	    fflush( stderr );
	}

	if( pipepid != 0 && flag ) {
	    fputc( EOF, pipe_stdin );
	    fflush( pipe_stdin );
	    sleep( 1 );
	    Xclose_p();

	    fclose( pipe_stdin );
	    fclose( pipe_stdout );
	    fclose( pipe_stderr );
	}

	if( TRUE ) {
	    fprintf( stderr, "leave pipe_cleanup[pid=%d]\n", pipepid );
	    fflush( stderr );
	}
}

void child_dead( sig )
	int sig;
{
	int pid;

	/* get pid */
	pid = wait( NULL );
	if( pid == pipepid ) {
	    if( TRUE ) {
		fprintf( stderr, "in child_dead [pid=%d]\n", pid );
		fflush( stderr );
	    }
	    signal( SIGCHLD, child_dead );
	    pipe_cleanup( FALSE );
	    pipepid = 0;
	}

	else{
	    fprintf( stderr, "comm process died!\n" );
	    catcher( SIGCHLD, 1 );
	}
}


/*  treat substr as a leading substring of str;
 *  return the number of chars matched.  If that is not the
 *  full length of str, then the next char must be a space or tab.
 */

int match( str, substr )
	unsigned char *str, *substr;
{
	int i;

	for( i=0; substr[i] != '\0' && str[i] == substr[i]; ++i );

	if( i < (int)strlen( (char *)str ) &&
	    ( str[i] != ' ' && str[i] != '\t' )) i = 0;

	if( FALSE ) fprintf( stderr, "MATCH:  return %d\n", i );

	return( i );
}

int iscodefile( c, s )
	MessageType c;
	unsigned char *s;
{
	MessageType i;

	if( c == HELP || c == STORED || c == INFO || c == THIST ||
	    c == LAST || c == REVIEW ||
	    c == PROVERB || c == MESSAGE || c == FIL ) {

	    if( s == NULL ) return( TRUE );
	    else if( sscanf( (char *)s, "%d File", &i ) == 1 && i == c )
	    	return( TRUE );
	    
	}
	return( FALSE );
}

get_game_info( g )
	int g;		/* game number to get info */
{
	if( dbg ) fprintf( stderr, "get games info [%d]\n", g );

	rank1[0] = rank2[0] = '\0';
	game_info_flag = g;

	fprintf( ffp, "games %d\n", g );
	fflush( ffp );
	sleep( 2 );
}

/* who.c
 * Elton Huang	5/93
 *
 * filter the who output so that it is sorted according to rank
 */

/*  Copyright (c) 1993 by Elton Huang.  All rights reserved.
 *
 *  This program is distributed in the hope that it will be useful.
 *  Use and copying of this software and preparation of derivative works
 *  based upon this software are permitted, so long as the following
 *  conditions are met:
 *       o credit to the authors is acknowledged following current
 *         academic behavior
 *       o no fees or compensation are charged for use, copies, or
 *         access to this software
 *       o this copyright notice is included intact.
 *  This software is made available AS IS, and no warranty is made about 
 *  the software or its performance. 
 * 
 *  Bug descriptions, use reports, comments or suggestions are welcome.
 *  Send them to    elton@cat.syr.edu
 */

/* this file is to provide the sorting by rank capability for who command for
   xigc.  To use it replace line 396 of filter.c with "who (current_str);",
   recomple filter.c and this file and link the object codes of this file
   with the rest of xigc object files to generate the new xigc. */

/* Yes, it is okay with me to include in the xigc release prividing that, just
a formality, the credit context is intact.  Please do let me if there is
any non-portablity to other UNIX variants.  Thanks,  Elton */

/* slightly hacked for xigc v3.8      =SC 6/95 */

struct bele {
    char str [40];
    struct bele *next;
};

static char new_who = 1;
static char do_rank();

#define BUCKET_SIZE (30 /*k*/ + 9 /*d*/ + 9 /*p*/ + 5)*2

struct bele *buckets [BUCKET_SIZE] = { NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL };

static void who (str)
    char *str;
{
    struct bele *player [2];
    int r;
    char s[120];

    if (new_who) { /* banner */
	putout( str, TRUE, TRUE, 141 );
	new_who = 0;
    } else
	/* players line; presume seperation bar exists for the last line
	 * even when the number of player is odd
	 */
	if (str [35] == '|') {
	    player [0] = (struct bele *) malloc (sizeof (struct bele));
	    if (player [0] == NULL) {
		fprintf (stderr, "xigc: Out of memory.\n");
		Quit( NULL, NULL, NULL );
	    }
	    strncpy (player [0] -> str, str, 34);
	    player [0] -> str [34] = 0;
	    player [0] -> next = buckets [r = do_rank (player [0] -> str)];
	    buckets [r] = player [0];
	    if (str [42] == '-' || ('0' <= str [42] && str [42] <= '9')) {
		/* to handle odd last line */
		player [1] = (struct bele *) malloc (sizeof (struct bele));
		if (player [1] == NULL) {
		    fprintf (stderr, "xigc: Out of memory.\n");
		    Quit( NULL, NULL, NULL );
		}
		strncpy (player [1] -> str, &str[37], 34);
		player [1] -> str [34] = 0;
		player [1] -> next = buckets [r = do_rank (player [1] -> str)];
		buckets [r] = player [1];
	    }
	} else { /* last line, dump out and free mem */
	    int i, rite;
	    struct bele *bi, *tmp;

#ifdef REVERSE
	    for (rite = 0, i = BUCKET_SIZE - 1; i >= 0; i--) {
#else
	    for (i = rite = 0; i < BUCKET_SIZE; i++) {
#endif
		for (bi = buckets [i]; bi;) {
		    sprintf (s, "%s%s", bi -> str, 
			     rite ? "" : " | ");
		    putout( s, TRUE, rite, 142 );
		    rite = !rite;
		    tmp = bi;
		    bi = bi -> next;
		    free (tmp);
		}
		buckets [i] = (struct bele *) NULL;
	    }
	    putout( "", TRUE, TRUE, 143 );
	    putout( str, TRUE, TRUE, 144 );
	    new_who = 1;
	}
}


static char do_rank (ss)
    char *ss;
{
    char buf [3];
    int num, base, fff;

    buf [2] = 0;
    fff = 0;

    if( ss [31] == 'N' ) return( 0 );
    else if( ss [31] == '?' ) return( 1 );
    else if( num  = atoi( strncpy(buf, ss + 30, 2) ) ) {
	if( ss [32] == 'k' ) base = -1; else base = 1;
	if( ss [32] == 'p' ) { base = 1; fff += 10; }

	return( base*num + 31 + fff ); 
    }

    else return( 0 );
}

/* end of who.c */


/* returns the number of items filled (should be 12)
 *
 * If IGS output was formatted better, this complex
 * function would not be needed =SC
 */
parse_games( str, game_descp )

	char *str;	/* string to parse */
	struct game_desc *game_descp;
{
	int ret=0, i;
	char *ret1, *ret2;
	unsigned char lstr[200];
	unsigned char *ls;

	strcpy( (char *)lstr, (char *)str );

	ls = lstr;

	/* get game number */
	for( ; *ls == ' ' || *ls == '\t' || *ls == '[' ; ++ls );

	if( *ls == '\0' || *ls == '#') return( ret );
	ret = sscanf( (char *)ls, "%d", &game_descp->game_no );

	for( ; *ls != ']' && *ls != '\0'; ++ls );
	if( *ls == '\0' ) return( ret );
	++ls;

	/* get white player */
	for( ; *ls == ' ' || *ls == '\t'; ++ls );
	if( *ls == '\0' ) return( ret );
	ret += sscanf( (char *)ls, "%s", game_descp->white_name );
	ls += strlen(game_descp->white_name );

	/* skip past '[' */
	for( ; *ls == ' ' || *ls == '\t' || *ls == '[' ; ++ls );
	if( *ls == '\0' ) return( ret );

	/* get white rank */
	for( i = 0 ; *ls != ' ' && *ls != '\t' && *ls != ']' ; ++ls ) {
	    if( *ls == '\0' ) return( ret );
	    game_descp->white_rank[i++] = *ls;
	}
	game_descp->white_rank[i] = '\0';
	game_descp->white_num_rank = convert_rank(game_descp->white_rank);
	++ret;

	for( ; *ls == ' ' || *ls == '\t' || *ls == ']' ; ++ls );
	if( *ls == '\0' ) return( ret );

	/* get black player */
	ret1 = strtok( (char *)ls, " \t" );
	if( ret1 == NULL ) return( ret );
	ret1 = strtok( NULL, " \t" );
	if( ret1 == NULL ) return( ret );
	strcpy( game_descp->black_name, ret1 );
	++ret;

	ret1 = strtok( NULL, "\0" );

	/* skip past '[' */
	for( ; *ret1 == ' ' || *ret1 == '\t' || *ret1 == '[' ; ++ret1 );
	if( *ret1 == '\0' ) return( ret );

	/* get black rank */
	for( i = 0 ; *ret1 != ' ' && *ret1 != '\t' && *ret1 != ']' ; ++ret1 ) {
	    if( *ret1 == '\0' ) return( ret );
	    game_descp->black_rank[i++] = *ret1;
	}
	game_descp->black_rank[i] = '\0';
	game_descp->black_num_rank = convert_rank(game_descp->black_rank);
	++ret;

	for( ; *ret1 == ' ' || *ret1 == '\t' ||
	       *ret1 == ']' || *ret1 == '(' ; ++ret1 );
	if( *ret1 == '\0' ) return( ret );

	/* get move, size, hcap, komi, byoyomi */
	ret += sscanf( ret1, "%d %d %d %f %d",
		&game_descp->move_num, &game_descp->board_size, &game_descp->handicap,
		&game_descp->komi, &game_descp->byo_yomi_min);
	for( ; *ret1 != ')' && *ret1 != '\0'; ++ret1 );
	if( *ret1 == '\0' ) return( ret );

	/* non-rated (free) game? */
	game_descp->free = *(ret1-2);
	++ret;
	game_descp->rules = *(ret1-1);

	/* get number of observers */
	for( ; *ret1 != '(' && *ret1 != '\0'; ++ret1 );
	if( *ret1 == '\0' ) return( ret );
	++ret1;
	ret += sscanf( ret1, "%d", &game_descp->ob_cnt );
	
	return( ret );
}

/* process sgf inside the "review" File blocks */
do_review( str, flag )
	char *str;	/* the line */
	int flag;	/* TRUE if newline, FALSE if no newline */
{
	static char s[LINESIZE];
	static int len;

	if( dbg ) {
	    fprintf( stderr, "do_review [0x%x] >>%s<<\n", flag, str );
	}

	strcat( s, str );
	len = strlen( s );

	if( flag ) {
	    sgf_command( s,flag );
	    s[0] = '\0';
	}

	return( 0 );
}


/* The following routines that begin "sgf_" were contributed by 
   James R. Logan Jr. and are in the public domain.  These routines 
   may be used by anyone for any purpose. */

#define SGF_IDLE 0
#define SGF_COMMENT 1
#define SGF_BNAME 2
#define SGF_WNAME 3
#define SGF_BRANK 4
#define SGF_WRANK 5
#define SGF_HANDICAP 6
#define SGF_KOMI 7
#define SGF_RESULT 8
#define SGF_DATE 9
#define SGF_SYSTEM 10
#define SGF_GAME 11
#define SGF_VIEW 12
#define SGF_WRATING 13
#define SGF_BRATING 14
#define SGF_SIZE 15
#define SGF_EVENT 16
#define SGF_USER 17
#define SGF_REGION 18
#define SGF_VERSION 19

sgf_command( s,flag )
	char *s;
	int flag;
{
	char *sgf_data = s;
	int done = FALSE;
	int x,y;

	if (sgf_lookfor) {
	    while ((*s != sgf_lookfor || (s != sgf_data && *(s-1) == '\\'))
					&& (*s != '\0'))
	        s++;
	    if (*s == sgf_lookfor) 
	        sgf_lookfor = 0;
	}

	while (!done) {
	    switch (*s++) {
		case 'A':			/* add/erase stone(s) */
		    if (sgf_action == SGF_IDLE) { 
	                switch (*s++) {
		            case 'B':		/* add black stone(s) */
	                        switch (*s) {
		                    case '[':	/* [move] */
				        while (*s == '[') { 
					    if ((*(s+1)-'a'>=maingame.width) || 
						(*(s+2)-'a'>=maingame.width)) {
					      ++s;
		            	              makemove( (int)(*s - 'a' + 1), 
					        19 - (int)(*(s+1) - 'a'), 0,
					        EMPTY, REALMOVE,
					        !boardchanged && !mouseflag,
					        ((rules==IRULES)
						?RULES:GOE_RULES));
					      ++s;
					    }
					    else {
					      ++s;
					      if( !isastone( (int)(*s-'a'+1),
							19-(int)(*(s+1) -'a'),
							REALMOVE ) ) {
		            	              makemove( (int)(*s - 'a' + 1), 
					        19 - (int)(*(s+1) - 'a'), 0,
					        BLACK, REALMOVE,
					        !boardchanged && !mouseflag,
					        ((rules==IRULES)
						?RULES:GOE_RULES));
					      }
					      ++s;
					    }
					    s += 2;
				        }
			                break;
		                    default:
				        sgf_lookfor = ']';
			                break;
		                }
			        break;
		            case 'E':		/* erase (lift) stone(s) */
	                        switch (*s) {
		                    case '[':	/* [move] */
				        while (*s == '[') { 
					    ++s;
					    x = *s - 'a' + 1;
					    y = 19 - (*(s+1) - 'a');
				            liftastone(&maingame,
					      y-1+(maingame.width*(x-1)),
					      REALGAME);
					    s += 3;
				        }
			                break;
		                    default:
				        sgf_lookfor = ']';
			                break;
		                }
			        break;
		            case 'W':		/* add white stone(s) */
	                        switch (*s) {
		                    case '[':	/* [move] */
				        while (*s == '[') { 
					    if ((*(s+1)-'a'>=maingame.width) || 
						(*(s+2)-'a'>=maingame.width)) {
					      ++s;
		            	              makemove( (int)(*s - 'a' + 1), 
					        19 - (int)(*(s+1) - 'a'), 0,
					        EMPTY, REALMOVE,
					        !boardchanged && !mouseflag,
					        ((rules==IRULES)
						?RULES:GOE_RULES));
					      ++s;
					    }
					    else {
					      ++s;
					      if( !isastone( (int)(*s-'a'+1),
							19-(int)(*(s+1) -'a'),
							REALMOVE ) ) {
		            	              makemove( (int)(*s - 'a' + 1), 
					        19 - (int)(*(s+1) - 'a'), 0,
					        WHITE, REALMOVE,
					        !boardchanged && !mouseflag,
					        ((rules==IRULES)
						?RULES:GOE_RULES));
					      }
					      ++s;
					    }
					    s += 2;
				        }
			                break;
		                    default:
				        sgf_lookfor = ']';
			                break;
		                }
			        break;
		            default:
			        sgf_lookfor = ']';
			        break;
		    }   }
		    break;
		case 'B':			/* black */
		    if (sgf_action == SGF_IDLE) { 
	                switch (*s++) {
		            case '[':		/* move */
				if ((*(s+1)-'a'>=maingame.width) || 
				    (*(s+2)-'a'>=maingame.width)) {
		       	            makemove( (int)(*s - 'a' + 1), 
				        19 - (int)(*(s+1) - 'a'), 0,
				        EMPTY, REALMOVE,
				        !boardchanged && !mouseflag,
				        ((rules==IRULES)?RULES:GOE_RULES));
				    s += 2;
	                            fprintf(stderr,"Pass\n");
				} else {
		       	            makemove( (int)(*s - 'a' + 1), 
				        19 - (int)(*(s+1) - 'a'), 0,
				        BLACK, REALMOVE,
				        !boardchanged && !mouseflag,
				        ((rules==IRULES)?RULES:GOE_RULES));
				    s += 2;
				}
			        sgf_lookfor = ']';
		                break;
		            case 'L':		/*  */
			        sgf_lookfor = ']';
			        break;
		            case 'R':		/* rank */
		    	        sgf_data = ++s;
		                sgf_action = SGF_BRANK;
			        sgf_lookfor = ']';
			        break;
		            default:
			        sgf_lookfor = ']';
			        break;
		    }   }
		    break;
		case 'C':			/* comment */
		    if (sgf_action == SGF_IDLE) { 
		        sgf_action = SGF_COMMENT;
		        sgf_data = ++s;
		        sgf_lookfor = ']';
		        fprintf(stderr,"\n");
		    }
		    break;
		case 'V':			/* view */
		    if (sgf_action == SGF_IDLE) {
			sgf_data = s + 2;
			sgf_action = SGF_VIEW;
			sgf_lookfor = ']';
		    }
		    break;
		case 'D':			/* date */
		    if (sgf_action == SGF_IDLE) {
			sgf_data = s + 2;
			sgf_action = SGF_DATE;
			sgf_lookfor = ']';
		    }
		    break;
		case 'E':			/* event */
		    if (sgf_action == SGF_IDLE) {
			sgf_data = s + 2;
			sgf_action = SGF_EVENT;
			sgf_lookfor = ']';
		    }
		    break;
		case 'G':			/* game */
		    if (sgf_action == SGF_IDLE) {
			sgf_data = s + 2;
			sgf_action = SGF_GAME;
			sgf_lookfor = ']';
		    }
		    break;
		case 'H':			/* handicap */
		    if (sgf_action == SGF_IDLE) { 
		        sgf_data = s + 2;
		        sgf_action = SGF_HANDICAP;
		        sgf_lookfor = ']';
		    }
		    break;
		case 'K':			/* komi */
		    if (sgf_action == SGF_IDLE) {
			if (*s == 'M') {
		            sgf_data = s + 2;
		            sgf_action = SGF_KOMI;
		            sgf_lookfor = ']';
			}
			else if (*s == 'I') {
			    sgf_data = (s += 2);
			    sgf_lookfor = ']';
			}
		    }
		    break;
		case 'N': /* Numerical rank */
			if (sgf_action == SGF_IDLE) {
				if (*s == 'W')
					sgf_action = SGF_WRATING;
				else if (*s == 'B')
					sgf_action = SGF_BRATING;
				sgf_data = (s += 2);
				sgf_lookfor = ']';
			}
			break;

		case 'P':			/* player */
		    if (sgf_action == SGF_IDLE) { 
	                switch (*s) {
		            case 'B':		/* login name of B */
		    	        sgf_data = s + 2;
		                sgf_action = SGF_BNAME;
			        break;
		            case 'C':		/* name of system */
		                if (sgf_action == SGF_IDLE) {
			            sgf_data = s + 2;
			            sgf_action = SGF_SYSTEM;
			            sgf_lookfor = ']';
		                }
			        break;
		            case 'W':		/* login name of W */
		    	        sgf_data = s + 2;
		                sgf_action = SGF_WNAME;
			        break;
		        }
		        sgf_lookfor = ']';
		    }
		    break;
		case 'R':			/* result */
		    if (sgf_action == SGF_IDLE) {
	            switch (*s++) {
		            case 'E':		/* result */
						sgf_data = s + 1;
						sgf_action = SGF_RESULT;
						sgf_lookfor = ']';
						break;
		            case 'G':  /* region for marking */

						fprintf( stderr, "MARKED REGION: " );
				        /*XXX  just ignore all these marked points */
				        while (*s == '[') {
					    	fprintf( stderr, "[%c%c]", *(s+1), *(s+2) );
					    	s += 4;
				        }
						fprintf( stderr, "\n" );
			        	break;
					default:
						break;
		        }
		    }
		    break;
		case 'S':			/* board size and system */
		    if (sgf_action == SGF_IDLE) {
	            switch (*s++) {
		            case 'Z':		/* size */
						sgf_data = s + 1;
						sgf_action = SGF_SIZE;
						break;
		            case 'Y':		/* system */
						sgf_data = s + 1;
						sgf_action = SGF_SYSTEM;
						break;
		        }
			sgf_lookfor = ']';
		    }
		    break;
		case 'F':			/* SGF version (file format) */
		    if (sgf_action == SGF_IDLE) {
			sgf_data = s + 2;
			sgf_action = SGF_VERSION;
			sgf_lookfor = ']';
		    }
		    break;
		case 'U':			/* user */
		    if (sgf_action == SGF_IDLE) {
			sgf_data = s + 2;
			sgf_action = SGF_USER;
			sgf_lookfor = ']';
		    }
		    break;
		case 'W':			/* white */
		    if (sgf_action == SGF_IDLE) {
	                switch (*s++) {
		            case '[':		/* move */
				if ((*(s+1)-'a'>=maingame.width) || 
				    (*(s+2)-'a'>=maingame.width)) {
		       	            makemove( (int)(*s - 'a' + 1), 
				        19 - (int)(*(s+1) - 'a'), 0,
				        EMPTY, REALMOVE,
				        !boardchanged && !mouseflag,
				        ((rules==IRULES)?RULES:GOE_RULES));
				    s +=2;
	                            fprintf(stderr,"Pass\n");
				} else {
		       	            makemove( (int)(*s - 'a' + 1), 
				        19 - (int)(*(s+1) - 'a'), 0,
				        WHITE, REALMOVE,
				        !boardchanged && !mouseflag,
				        ((rules==IRULES)?RULES:GOE_RULES));
				     s +=2;
				}
			        sgf_lookfor = ']';
		                break;
		            case 'L':		/* luck factor */
			        sgf_lookfor = ']';
			        break;
		            case 'R':		/* rank */
		    	        sgf_data = s + 1;
		                sgf_action = SGF_WRANK;
			        sgf_lookfor = ']';
			        break;
		            default:
			        sgf_lookfor = ']';
			        break;
		    }   }
		    break;
		case '(':			/* begin variation */
		    if (sgf_action == SGF_IDLE) 
		        sgf_var_begin[sgf_var_depth++] = maingame.stoneNumber;
		    break;
		case ')':			/* end variation */
		    if (sgf_action == SGF_IDLE) {
		        if (sgf_var_depth > 0) {
		            sgf_var_depth--;
		            while (maingame.stoneNumber > 
				   sgf_var_begin[sgf_var_depth]) {
	                        backup(&maingame,maingame.stoneNumber,DRAWIT,
		        	       ((rules==IRULES)?RULES:GOE_RULES));
				maingame.stoneNumber--;
		        }   }
	  	        backupend(&maingame,
				  sgf_var_begin[sgf_var_depth],
				  DRAWIT,
				  0);
		    }
		    break;
		case '[':			/* begin data */
		    if (sgf_action == SGF_IDLE) 
		        sgf_lookfor = ']';
		    break;
		case 0:				/* end of line */
		    done = TRUE;
		    sgf_display(s,sgf_data,flag);
		    break;
		case ']':			/* end data */
		    sgf_display(s,sgf_data,flag);
		    sgf_action = SGF_IDLE;
		    if (sgf_lookfor == ']') sgf_lookfor = 0;
		    break;
		case ';':			/*  */
		    break;
		default:
	            break;
	    }
	    if (sgf_lookfor) {
	        while ((*s != sgf_lookfor || (s != sgf_data && *(s-1) == '\\'))
					&& (*s != '\0'))
	            s++;
	        if (*s == sgf_lookfor ) {
	            sgf_lookfor = 0;		/* stop looking */
	    	}
	    }
	}
}


sgf_display( s,sgf_data,flag )
	char *s;
	char *sgf_data;
	int flag;
{
    switch (sgf_action) {
	case SGF_COMMENT:
	    if (s > (sgf_data + 2)) {
	        *(s-1) = 0;
		if( flag ) putout( sgf_data, TRUE, flag, 123 );
		else  putout( sgf_data, TRUE, flag, 1 );
	    }
	    break;
	case SGF_BNAME:
	    *(s-1) = 0;
	    strcpy( (char *)player2,sgf_data);
	    update_status( gameno, rules, ffre,
		    player2,
		    player1, p2_captured, p1_captured,
		    p2_sec, p1_sec, p2_byoyomi,
		    p1_byoyomi, rank1, rank2 );
	    fprintf(stderr,"Black Player: %s\n",sgf_data);
	    break;
	case SGF_WNAME:
	    *(s-1) = 0;
	    strcpy( (char *)player1,sgf_data);
	    update_status( gameno, rules, ffre,
		    player2,
		    player1, p2_captured, p1_captured,
		    p2_sec, p1_sec, p2_byoyomi,
		    p1_byoyomi, rank1, rank2 );
	    fprintf(stderr,"White Player: %s\n",sgf_data);
	    break;
	case SGF_BRANK:
	    *(s-1) = 0;
	    strcpy( (char *)rank2,sgf_data);
	    update_status( gameno, rules, ffre,
		    player2,
		    player1, p2_captured, p1_captured,
		    p2_sec, p1_sec, p2_byoyomi,
		    p1_byoyomi, rank1, rank2 );
	    fprintf(stderr,"Black Rank: %s\n",sgf_data);
	    break;
	case SGF_WRANK:
	    *(s-1) = 0;
	    strcpy( (char *)rank1,sgf_data);
	    update_status( gameno, rules, ffre,
		    player2,
		    player1, p2_captured, p1_captured,
		    p2_sec, p1_sec, p2_byoyomi,
		    p1_byoyomi, rank1, rank2 );
	    fprintf(stderr,"White Rank: %s\n",sgf_data);
	    break;
	case SGF_HANDICAP:
	    *(s-1) = 0;
	    sscanf(sgf_data,"%d",&handicap);
	    update_handicap( handicap, komi );
	    break;
	case SGF_KOMI:
	    *(s-1) = 0;
	    sscanf(sgf_data,"%f",&komi);
	    fprintf(stderr,"Komi: %4.1f\n", komi );
	    update_handicap( handicap, komi );
	    break;
	case SGF_VERSION:
	    *(s-1) = 0;
	    sscanf( sgf_data,"%d",&sgf_version );
	    if( sgf_version > 3 ) {
	    	 fprintf( stderr,
	    	 	"\n\nERROR: this version of xigc does not support SGF version %d\n",
	    	 		 sgf_version );
	    	 fprintf( stderr,
	    	 	"ERROR: results beyond this point are unpredictable!\n\n" );
	    	 beep();
		}
	    else fprintf( stderr, "SGF Version: %d\n", sgf_version );
	    break;
	case SGF_USER:
	    *(s-1) = 0;
	    fprintf(stderr,"User: %s\n",sgf_data);
	    break;
	case SGF_RESULT:
	    *(s-1) = 0;
	    fprintf(stderr,"Result: %s\n",sgf_data);
	    break;
	case SGF_SIZE:		/* should set board size if not 19, but
				 * SGF apparently does not support other
				 * board sizes.... 9x9 games are in upper
				 * left corner of board
				 */
	    *(s-1) = 0;
	    fprintf(stderr,"Board Size: %s\n",sgf_data);
	    break;
	case SGF_DATE:
	    *(s-1) = 0;
	    fprintf(stderr,"Date: %s\n",sgf_data);
	    break;
	case SGF_EVENT:
	    *(s-1) = 0;
	    fprintf(stderr,"Event: %s\n",sgf_data);
	    break;
	case SGF_SYSTEM:
	    *(s-1) = 0;
	    fprintf(stderr,"System: %s\n",sgf_data);
	    break;
	case SGF_GAME:
	    *(s-1) = 0;
	    fprintf(stderr,"Game: %s\n",sgf_data);
	    break;
	case SGF_VIEW:
	    *(s-1) = 0;
	    fprintf(stderr,"View: %s\n",sgf_data);
	    break;
	case SGF_WRATING:
		*(s-1) = 0;
		fprintf(stderr,"White rating: %s\n", sgf_data);
		break;
	case SGF_BRATING:
		*(s-1) = 0;
		fprintf(stderr,"Black rating: %s\n", sgf_data);
		break;		
	case SGF_REGION:	/* ignore this for the moment */
		break;
	default:
	    break;
    }
}


sgf_clear_info() {
	gameno = 0;
	rules = 'I';
	ffre = ' ';
	player1[0] = 0;
	player2[0] = 0;
	p2_captured = 0;
	p1_captured = 0;
	p2_sec = 0;
	p1_sec = 0;
	p2_byoyomi = 0;
	p1_byoyomi = 0;
	rank1[0] = 0;
	rank2[0] = 0;

	handicap = 0;
	komi = .5;
}

int process_local_sgf_line( f )
	int f;		/* flag is 0 to reset, 1 to read a line */
{
	static char str[10000];
	static char instr[1000];
	static int i, len=0;
	static int ip=0, sp=0;
	static unsigned char inblockflag = FALSE, beenhere = FALSE;

	if( dbg ) fprintf( stderr, "Enter process_sgf_file\n" );

	if( f == 0 ) {
	    i = len = ip = sp = 0;
	    inblockflag = beenhere = FALSE;
	    return( 0 );
	}

    more_lines:

	if( ip >= len-1 ) {
	    if( fgets( (char *)instr, 1000, sgf_file_fp ) == NULL ) {
	        fclose( sgf_file_fp );
	        sgflocalflag = FALSE;
	        fprintf( stderr, "LOCAL: End of sgf file\n" );
	        return( (-1) );
	    }
	    if( dbg ) fprintf( stderr, "\tLINE: %s\n", instr );
	    len = strlen( instr );
	    ip = 0;
	}

	/* we do a little bit of pre-processing here, to pass only
	 * terse-form sgf to the handler function
	 */

	for( ; ip < len-1;  ++ip ) {
	    if( instr[ip] == '\0' ) {
	    }
	    else if( instr[ip] == '[' ) {
		inblockflag = TRUE;
		if( beenhere ) str[sp++] = instr[ip];
	    }
	    else if( instr[ip] == ']' && instr[ip-1] != '\\' ) {
		inblockflag = FALSE;
		if( beenhere ) str[sp++] = instr[ip];
	    }
/*XXX must strip newlines if inblockflag == TRUE and !comment */
	    else if( inblockflag ) {
		if( beenhere ) str[sp++] = instr[ip];
	    }
	    else if( isalpha( instr[ip] ) && isupper( instr[ip] ) ) {
		if( beenhere ) str[sp++] = instr[ip];
	    }
	    else if( instr[ip] == ';' ) {
		if( beenhere ) {
		    str[sp++] = instr[ip];
		    str[sp] = '\0';
		}
		else {
		    str[0] = '(';
		    str[1] = ';';
		    str[2] = '\0';
		    if( tersemode ) fprintf( stderr, "\r" );
		    else fprintf( stderr, "\n" );
		}
		do_review( str, TRUE );
		sp = 0;
		++ip;
		if( !beenhere ) {
		    beenhere = TRUE;
		    goto more_lines;
		}
		break;
	    }
	    else if( !isalpha( instr[ip] ) ) {
		if( beenhere ) str[sp++] = instr[ip];
	    }
	}

	if( !beenhere || ( sp != 0 && ip >= len-1 ) ) {
	    if( inblockflag ) str[sp++] = '\n';
	    goto more_lines;
	}
		
	if( dbg ) fprintf( stderr, "Leave process_sgf_file\n" );
	return( 0 );
}
